@@ -17,11 +17,13 @@
 ################################################################################
 
 from abc import ABCMeta
+from enum import Enum
 
 from py4j.java_gateway import get_java_class
+from typing import List, Optional
 
 from pyflink.java_gateway import get_gateway
-from pyflink.util.utils import load_java_class
+from pyflink.util.java_utils import load_java_class
 
 __all__ = [
     'StateBackend',
@@ -192,9 +194,8 @@ class MemoryStateBackend(StateBackend):
                                the runtime configuration will be used.
         :param max_state_size: The maximal size of the serialized state. If none, the
                                :data:`DEFAULT_MAX_STATE_SIZE` will be used.
-        :param using_asynchronous_snapshots: Flag to switch between synchronous and asynchronous
-                                             snapshot mode. If null, the value configured in the
-                                             runtime configuration will be used.
+        :param using_asynchronous_snapshots: Snapshots are now always asynchronous. This flag
+                                             has no effect anymore in this version.
         :param j_memory_state_backend: For internal use, please keep none.
         """
         if j_memory_state_backend is None:
@@ -222,7 +223,7 @@ class MemoryStateBackend(StateBackend):
         self._j_memory_state_backend = j_memory_state_backend
         super(MemoryStateBackend, self).__init__(j_memory_state_backend)
 
-    def get_max_state_size(self):
+    def get_max_state_size(self) -> int:
         """
         Gets the maximum size that an individual state can have, as configured in the
         constructor (by default :data:`DEFAULT_MAX_STATE_SIZE`).
@@ -231,7 +232,7 @@ class MemoryStateBackend(StateBackend):
         """
         return self._j_memory_state_backend.getMaxStateSize()
 
-    def is_using_asynchronous_snapshots(self):
+    def is_using_asynchronous_snapshots(self) -> bool:
         """
         Gets whether the key/value data structures are asynchronously snapshotted.
 
@@ -323,9 +324,8 @@ class FsStateBackend(StateBackend):
         :param write_buffer_size: Write buffer size used to serialize state. If -1, the value
                                   configured in the runtime configuration will be used, or the
                                   default value (4KB) if nothing is configured.
-        :param using_asynchronous_snapshots: Flag to switch between synchronous and asynchronous
-                                             snapshot mode. If none, the value configured in
-                                             the runtime configuration will be used.
+        :param using_asynchronous_snapshots: Snapshots are now always asynchronous. This flag
+                                             has no effect anymore in this version.
         :param j_fs_state_backend: For internal use, please keep none.
         """
         if j_fs_state_backend is None:
@@ -369,7 +369,7 @@ class FsStateBackend(StateBackend):
         self._j_fs_state_backend = j_fs_state_backend
         super(FsStateBackend, self).__init__(j_fs_state_backend)
 
-    def get_checkpoint_path(self):
+    def get_checkpoint_path(self) -> str:
         """
         Gets the base directory where all the checkpoints are stored.
         The job-specific checkpoint directory is created inside this directory.
@@ -378,7 +378,7 @@ class FsStateBackend(StateBackend):
         """
         return self._j_fs_state_backend.getCheckpointPath().toString()
 
-    def get_min_file_size_threshold(self):
+    def get_min_file_size_threshold(self) -> int:
         """
         Gets the threshold below which state is stored as part of the metadata, rather than in
         files. This threshold ensures that the backend does not create a large amount of very
@@ -391,7 +391,7 @@ class FsStateBackend(StateBackend):
         """
         return self._j_fs_state_backend.getMinFileSizeThreshold()
 
-    def is_using_asynchronous_snapshots(self):
+    def is_using_asynchronous_snapshots(self) -> bool:
         """
         Gets whether the key/value data structures are asynchronously snapshotted.
 
@@ -403,7 +403,7 @@ class FsStateBackend(StateBackend):
         """
         return self._j_fs_state_backend.isUsingAsynchronousSnapshots()
 
-    def get_write_buffer_size(self):
+    def get_write_buffer_size(self) -> int:
         """
         Gets the write buffer size for created checkpoint stream.
 
@@ -503,7 +503,7 @@ class RocksDBStateBackend(StateBackend):
         j_state_backend = self._j_rocks_db_state_backend.getCheckpointBackend()
         return _from_j_state_backend(j_state_backend)
 
-    def set_db_storage_paths(self, *paths):
+    def set_db_storage_paths(self, *paths: str):
         """
         Sets the directories in which the local RocksDB database puts its files (like SST and
         metadata files). These directories do not need to be persistent, they can be ephemeral,
@@ -531,7 +531,7 @@ class RocksDBStateBackend(StateBackend):
                 j_path_array[i] = paths[i]
             self._j_rocks_db_state_backend.setDbStoragePaths(j_path_array)
 
-    def get_db_storage_paths(self):
+    def get_db_storage_paths(self) -> List[str]:
         """
         Gets the configured local DB storage paths, or null, if none were configured.
 
@@ -547,7 +547,7 @@ class RocksDBStateBackend(StateBackend):
         """
         return list(self._j_rocks_db_state_backend.getDbStoragePaths())
 
-    def is_incremental_checkpoints_enabled(self):
+    def is_incremental_checkpoints_enabled(self) -> bool:
         """
         Gets whether incremental checkpoints are enabled for this state backend.
 
@@ -555,7 +555,7 @@ class RocksDBStateBackend(StateBackend):
         """
         return self._j_rocks_db_state_backend.isIncrementalCheckpointsEnabled()
 
-    def set_predefined_options(self, options):
+    def set_predefined_options(self, options: 'PredefinedOptions'):
         """
         Sets the predefined options for RocksDB.
 
@@ -571,26 +571,9 @@ class RocksDBStateBackend(StateBackend):
 
         :param options: The options to set (must not be null), see :class:`PredefinedOptions`.
         """
-        gateway = get_gateway()
-        JPredefinedOptions = gateway.jvm.org.apache.flink.contrib.streaming.state.PredefinedOptions
-        if options == PredefinedOptions.DEFAULT:
-            self._j_rocks_db_state_backend.setPredefinedOptions(JPredefinedOptions.DEFAULT)
-        elif options == PredefinedOptions.SPINNING_DISK_OPTIMIZED:
-            self._j_rocks_db_state_backend.setPredefinedOptions(
-                JPredefinedOptions.SPINNING_DISK_OPTIMIZED)
-        elif options == PredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM:
-            self._j_rocks_db_state_backend.setPredefinedOptions(
-                JPredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM)
-        elif options == PredefinedOptions.FLASH_SSD_OPTIMIZED:
-            self._j_rocks_db_state_backend.setPredefinedOptions(
-                JPredefinedOptions.FLASH_SSD_OPTIMIZED)
-        else:
-            raise TypeError("Unsupported options: %s, the supported options are: "
-                            "PredefinedOptions.DEFAULT, PredefinedOptions.SPINNING_DISK_OPTIMIZED,"
-                            " PredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM and "
-                            "PredefinedOptions.FLASH_SSD_OPTIMIZED")
+        self._j_rocks_db_state_backend.setPredefinedOptions(options._to_j_predefined_options())
 
-    def get_predefined_options(self):
+    def get_predefined_options(self) -> 'PredefinedOptions':
         """
         Gets the current predefined options for RocksDB.
         The default options (if nothing was set via :func:`setPredefinedOptions`)
@@ -606,20 +589,9 @@ class RocksDBStateBackend(StateBackend):
         :return: Current predefined options.
         """
         j_predefined_options = self._j_rocks_db_state_backend.getPredefinedOptions()
-        gateway = get_gateway()
-        JPredefinedOptions = gateway.jvm.org.apache.flink.contrib.streaming.state.PredefinedOptions
-        if j_predefined_options == JPredefinedOptions.DEFAULT:
-            return PredefinedOptions.DEFAULT
-        elif j_predefined_options == JPredefinedOptions.FLASH_SSD_OPTIMIZED:
-            return PredefinedOptions.FLASH_SSD_OPTIMIZED
-        elif j_predefined_options == JPredefinedOptions.SPINNING_DISK_OPTIMIZED:
-            return PredefinedOptions.SPINNING_DISK_OPTIMIZED
-        elif j_predefined_options == JPredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM:
-            return PredefinedOptions.SPINNING_DISK_OPTIMIZED_HIGH_MEM
-        else:
-            raise Exception("Unsupported java options: %s" % j_predefined_options)
+        return PredefinedOptions._from_j_predefined_options(j_predefined_options)
 
-    def set_options(self, options_factory_class_name):
+    def set_options(self, options_factory_class_name: str):
         """
         Sets ``org.rocksdb.Options`` for the RocksDB instances.
         Because the options are not serializable and hold native code references,
@@ -641,7 +613,7 @@ class RocksDBStateBackend(StateBackend):
             raise ValueError("The input class does not implement RocksDBOptionsFactory.")
         self._j_rocks_db_state_backend.setRocksDBOptions(j_options_factory_clz.newInstance())
 
-    def get_options(self):
+    def get_options(self) -> Optional[str]:
         """
         Gets the fully-qualified class name of the options factory in Java that lazily creates
         the RocksDB options.
@@ -654,7 +626,7 @@ class RocksDBStateBackend(StateBackend):
         else:
             return None
 
-    def get_number_of_transfering_threads(self):
+    def get_number_of_transfering_threads(self) -> int:
         """
         Gets the number of threads used to transfer files while snapshotting/restoring.
 
@@ -662,7 +634,7 @@ class RocksDBStateBackend(StateBackend):
         """
         return self._j_rocks_db_state_backend.getNumberOfTransferingThreads()
 
-    def set_number_of_transfering_threads(self, number_of_transfering_threads):
+    def set_number_of_transfering_threads(self, number_of_transfering_threads: int):
         """
         Sets the number of threads used to transfer files while snapshotting/restoring.
 
@@ -675,7 +647,7 @@ class RocksDBStateBackend(StateBackend):
         return self._j_rocks_db_state_backend.toString()
 
 
-class PredefinedOptions(object):
+class PredefinedOptions(Enum):
     """
     The :class:`PredefinedOptions` are configuration settings for the :class:`RocksDBStateBackend`.
     The various pre-defined choices are configurations that have been empirically
@@ -765,10 +737,19 @@ class PredefinedOptions(object):
     SPINNING_DISK_OPTIMIZED_HIGH_MEM = 2
     FLASH_SSD_OPTIMIZED = 3
 
+    @staticmethod
+    def _from_j_predefined_options(j_predefined_options) -> 'PredefinedOptions':
+        return PredefinedOptions[j_predefined_options.name()]
+
+    def _to_j_predefined_options(self):
+        gateway = get_gateway()
+        JPredefinedOptions = gateway.jvm.org.apache.flink.contrib.streaming.state.PredefinedOptions
+        return getattr(JPredefinedOptions, self.name)
+
 
 class CustomStateBackend(StateBackend):
     """
-    A wrapper of customized java state backend created from the provided `StateBackendFactory`.
+    A wrapper of customized java state backend.
     """
 
     def __init__(self, j_custom_state_backend):
